import { WebApplicationEventManager, Logger, webApplication, WebApplicationEventType } from '@fly-vue/core';
import _ from 'lodash';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var AsyncModuleStatus;
(function (AsyncModuleStatus) {
    /** 未加载 */
    AsyncModuleStatus[AsyncModuleStatus["none"] = 0] = "none";
    /** 加载中 */
    AsyncModuleStatus[AsyncModuleStatus["loading"] = 1] = "loading";
    /** 已加载 */
    AsyncModuleStatus[AsyncModuleStatus["loaded"] = 2] = "loaded";
})(AsyncModuleStatus || (AsyncModuleStatus = {}));

var AsyncModuleRegistryEventType;
(function (AsyncModuleRegistryEventType) {
    /**
     * 加载模块之前（每个模块和所有依赖）
     */
    AsyncModuleRegistryEventType[AsyncModuleRegistryEventType["LoadModuleBefore"] = 1] = "LoadModuleBefore";
    /**
     * 加载模块成功之后（每个模块和所有依赖）
     */
    AsyncModuleRegistryEventType[AsyncModuleRegistryEventType["LoadModuleAfter"] = 2] = "LoadModuleAfter";
    /**
     * 加载模块之前（每个模块）
     */
    AsyncModuleRegistryEventType[AsyncModuleRegistryEventType["LoadModuleBeforeEach"] = 3] = "LoadModuleBeforeEach";
    /**
     * 加载模块成功之后（每个模块）
     */
    AsyncModuleRegistryEventType[AsyncModuleRegistryEventType["LoadModuleAfterEach"] = 4] = "LoadModuleAfterEach";
})(AsyncModuleRegistryEventType || (AsyncModuleRegistryEventType = {}));
var AsyncModuleRegistryEvents = new WebApplicationEventManager();

/**
 * 管理注册按需加载模块
 */
var AsyncModuleRegistry = /** @class */ (function () {
    function AsyncModuleRegistry() {
    }
    Object.defineProperty(AsyncModuleRegistry, "events", {
        /** 事件 */
        get: function () {
            return AsyncModuleRegistryEvents;
        },
        enumerable: false,
        configurable: true
    });
    AsyncModuleRegistry.find = function (name) {
        return _.find(AsyncModuleRegistry.modules, function (_module) { return _module.name === name; });
    };
    /**
     * 注册按需加载模块
     * @param moduleDef
     */
    AsyncModuleRegistry.register = function (moduleDef) {
        var _this = this;
        var _modules = _.isArray(moduleDef) ? moduleDef : [moduleDef];
        _.forEach(_modules, function (_module) {
            if (!_module.name) {
                Logger.error("\u5F02\u6B65\u6A21\u5757\u4E0D\u5408\u6CD5\uFF1A\u672A\u6307\u5B9Aname");
                return;
            }
            //未设置jsUrl的模块不能注册进来
            if (!_module.jsUrl) {
                Logger.error("\u5F02\u6B65\u6A21\u5757\u672A\u6307\u5B9AjsUrl\u5730\u5740\uFF1A" + _module.name);
                return;
            }
            if (!_module.publicPath) {
                var jsUrl = _module.jsUrl;
                _module.publicPath = jsUrl.substring(0, jsUrl.lastIndexOf('/modules')) + '/';
            }
            if (!_this.find(_module.name)) {
                _module.status = AsyncModuleStatus.none;
                AsyncModuleRegistry.modules.push(_module);
            }
        });
    };
    AsyncModuleRegistry.getAllModules = function () {
        return AsyncModuleRegistry.modules;
    };
    AsyncModuleRegistry.modules = [];
    return AsyncModuleRegistry;
}());

// function _makeVarName(name) {
//   return name.replace('@', '').replace(/[^a-z0-9]/gi, '_');
// }
//@fly-vue/yourproject-yourmodule => _$bingo_yourproject_yourmodule_
var _allcssUrls = [];
function AddCssByStyle(cssUrl) {
    if (_allcssUrls.indexOf(cssUrl) >= 0)
        return;
    _allcssUrls.push(cssUrl);
    var doc = document;
    var heads = doc.getElementsByTagName('head');
    var pEl = heads.length ? heads[0] : doc.documentElement;
    var link = document.createElement('link');
    link.type = 'text/css';
    link.rel = 'stylesheet';
    link.href = cssUrl;
    pEl.appendChild(link);
}
function _makeGlobalVarname(name) {
    return ['_$', name, '_'].join('');
}
function loadCss(cssUrl) {
    return __awaiter(this, void 0, void 0, function () {
        var errMsg;
        return __generator(this, function (_a) {
            errMsg = "\u52A0\u8F7D\u5F02\u6B65\u6A21\u5757css\u51FA\u9519:" + cssUrl;
            // const m = await window.System.import(cssUrl).catch((err) => {
            //   Messager.notice(errMsg);
            //   Logger.error(err);
            // });
            // if (!m) {
            //   return;
            // }
            try {
                // const styleSheet = m.default; // A CSSStyleSheet object
                // document.adoptedStyleSheets = [...document.adoptedStyleSheets, styleSheet];
                AddCssByStyle(cssUrl);
            }
            catch (err) {
                Logger.error(errMsg);
                Logger.error(err);
            }
            return [2 /*return*/];
        });
    });
}
function loadJs(moduleDef) {
    return __awaiter(this, void 0, void 0, function () {
        var jsUrl, dllModuleName, globalModuleId, moduleId, errMsg, isCancel, cancel, m, moduleInst, err_1;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    jsUrl = moduleDef.jsUrl;
                    dllModuleName = moduleDef.name;
                    globalModuleId = _makeGlobalVarname(dllModuleName);
                    moduleId = "./node_modules/" + dllModuleName + "/index.js";
                    errMsg = "\u52A0\u8F7D\u5F02\u6B65\u6A21\u5757js\u51FA\u9519:" + jsUrl;
                    isCancel = false;
                    cancel = function () {
                        isCancel = true;
                    };
                    return [4 /*yield*/, AsyncModuleRegistryEvents.dispatch(AsyncModuleRegistryEventType.LoadModuleBeforeEach, {
                            webApp: webApplication,
                            moduleDef: moduleDef,
                            cancel: cancel
                        })];
                case 1:
                    _a.sent();
                    if (isCancel)
                        return [2 /*return*/];
                    return [4 /*yield*/, window.System.import(jsUrl).catch(function (err) {
                            Logger.error(errMsg);
                            Logger.error(err);
                        })];
                case 2:
                    m = _a.sent();
                    if (!m) {
                        return [2 /*return*/];
                    }
                    _a.label = 3;
                case 3:
                    _a.trys.push([3, 7, , 8]);
                    if (window[globalModuleId]) {
                        window[globalModuleId](moduleId);
                    }
                    else {
                        m.default(moduleId);
                    }
                    moduleInst = webApplication.getModule(dllModuleName);
                    if (!moduleInst) return [3 /*break*/, 6];
                    // console.log(`loader ok：${dllModuleName}@${moduleInst.version}`)
                    //等待所有事件执行、路由添加等操作完成
                    return [4 /*yield*/, webApplication.installAsyncModule(moduleInst)];
                case 4:
                    // console.log(`loader ok：${dllModuleName}@${moduleInst.version}`)
                    //等待所有事件执行、路由添加等操作完成
                    _a.sent();
                    return [4 /*yield*/, AsyncModuleRegistryEvents.dispatch(AsyncModuleRegistryEventType.LoadModuleAfterEach, {
                            webApp: webApplication,
                            moduleDef: moduleDef,
                            module: moduleInst,
                            cancel: cancel
                        })];
                case 5:
                    _a.sent();
                    if (isCancel)
                        return [2 /*return*/];
                    _a.label = 6;
                case 6: return [3 /*break*/, 8];
                case 7:
                    err_1 = _a.sent();
                    Logger.error(errMsg);
                    Logger.error(err_1);
                    return [3 /*break*/, 8];
                case 8: return [2 /*return*/];
            }
        });
    });
}
/**
 * 加载依赖
 * @param moduleDef
 * @returns
 */
function loadDependencies(moduleDef) {
    return __awaiter(this, void 0, void 0, function () {
        var dependencies, waitAll;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    dependencies = moduleDef.dependencies;
                    if (_.size(dependencies) == 0)
                        return [2 /*return*/];
                    waitAll = _.map(dependencies, function (name) {
                        return __awaiter(this, void 0, void 0, function () {
                            var module, asyncModule;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        module = webApplication.getModule(name);
                                        if (module)
                                            return [2 /*return*/];
                                        asyncModule = AsyncModuleRegistry.find(name);
                                        if (!asyncModule)
                                            return [2 /*return*/];
                                        return [4 /*yield*/, loadModule(asyncModule)];
                                    case 1:
                                        _a.sent();
                                        return [2 /*return*/];
                                }
                            });
                        });
                    });
                    return [4 /*yield*/, Promise.all(waitAll)];
                case 1:
                    _a.sent();
                    return [2 /*return*/];
            }
        });
    });
}
/**
 * 等待加载中
 * @param moduleDef
 * @returns
 */
function waitLoading(moduleDef) {
    return new Promise(function (r) {
        var tick = 10, max = 120 * 1000; //120秒
        var count = 0;
        var id = setInterval(function () {
            if (moduleDef.status == AsyncModuleStatus.loaded) {
                clearInterval(id);
                r(true);
            }
            count += tick;
            if (count >= max) {
                clearInterval(id);
                r(false);
            }
        }, tick);
    });
}
/**
 * 异步加载dll模块
 * @param dllModuleName 模块名称
 * @returns
 */
var loadModule = function (moduleDef) { return __awaiter(void 0, void 0, void 0, function () {
    return __generator(this, function (_a) {
        switch (_a.label) {
            case 0:
                if (!(moduleDef.status == AsyncModuleStatus.loading)) return [3 /*break*/, 2];
                //加载中，防止重复加载，如：过快切换路由
                return [4 /*yield*/, waitLoading(moduleDef)];
            case 1:
                //加载中，防止重复加载，如：过快切换路由
                _a.sent();
                return [2 /*return*/];
            case 2:
                if (moduleDef.status != AsyncModuleStatus.none)
                    return [2 /*return*/];
                moduleDef.status = AsyncModuleStatus.loading;
                _a.label = 3;
            case 3:
                _a.trys.push([3, , 8, 9]);
                return [4 /*yield*/, loadDependencies(moduleDef)];
            case 4:
                _a.sent();
                return [4 /*yield*/, loadJs(moduleDef)];
            case 5:
                _a.sent();
                if (!moduleDef.cssUrl) return [3 /*break*/, 7];
                return [4 /*yield*/, loadCss(moduleDef.cssUrl)];
            case 6:
                _a.sent();
                _a.label = 7;
            case 7: return [3 /*break*/, 9];
            case 8:
                moduleDef.status = AsyncModuleStatus.loaded;
                return [7 /*endfinally*/];
            case 9: return [2 /*return*/];
        }
    });
}); };

function match(moduleDef, path) {
    var match = moduleDef.match;
    if (!match) {
        return false;
    }
    if (_.isString(match)) {
        return _.startsWith(path, match);
    }
    else {
        return match(path);
    }
}
function findMatchedModule(path) {
    var matchedModule;
    _.forEach(AsyncModuleRegistry.getAllModules(), function (moduleDef) {
        if (match(moduleDef, path)) {
            matchedModule = moduleDef;
            return false;
        }
    });
    return matchedModule;
}

var currentModule = {
    name: process.env.VUE_APP_NAME,
    version: process.env.VUE_APP_VERSION,
    dependencies: ['@fly-vue/core'],
    install: function () {
        webApplication.registerEvent(WebApplicationEventType.routerLoaded, function () {
            var router = webApplication.router;
            router.beforeEach(function (to, from, next) {
                return __awaiter(this, void 0, void 0, function () {
                    var matchedModule, exists, isCancel_1, cancel, path, query, hash;
                    return __generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                matchedModule = findMatchedModule(to.fullPath);
                                if (!matchedModule) return [3 /*break*/, 6];
                                exists = webApplication.getModule(matchedModule.name);
                                if (!!exists) return [3 /*break*/, 4];
                                isCancel_1 = false;
                                cancel = function () {
                                    isCancel_1 = true;
                                };
                                return [4 /*yield*/, AsyncModuleRegistryEvents.dispatch(AsyncModuleRegistryEventType.LoadModuleBefore, {
                                        webApp: webApplication,
                                        moduleDef: matchedModule,
                                        cancel: cancel,
                                        routerArgs: { to: to, from: from, next: next }
                                    })];
                            case 1:
                                _a.sent();
                                if (isCancel_1)
                                    return [2 /*return*/];
                                path = to.path, query = to.query, hash = to.hash;
                                return [4 /*yield*/, loadModule(matchedModule)];
                            case 2:
                                _a.sent();
                                exists = webApplication.getModule(matchedModule.name);
                                return [4 /*yield*/, AsyncModuleRegistryEvents.dispatch(AsyncModuleRegistryEventType.LoadModuleAfter, {
                                        webApp: webApplication,
                                        moduleDef: matchedModule,
                                        module: exists,
                                        cancel: cancel,
                                        routerArgs: { to: to, from: from, next: next }
                                    })];
                            case 3:
                                _a.sent();
                                if (isCancel_1)
                                    return [2 /*return*/];
                                if (exists) {
                                    next({ path: path, query: query, hash: hash });
                                }
                                else
                                    next(); //404
                                return [3 /*break*/, 5];
                            case 4:
                                next();
                                _a.label = 5;
                            case 5: return [3 /*break*/, 7];
                            case 6:
                                next();
                                _a.label = 7;
                            case 7: return [2 /*return*/];
                        }
                    });
                });
            });
        });
    }
};
webApplication.addModule(currentModule);

export { AsyncModuleRegistry, AsyncModuleRegistryEventType, AsyncModuleRegistryEvents, AsyncModuleStatus, loadModule };
